import argparse
import numpy as np
from ..gzip_utils import open_for_write
from ..dto.transcript_coverage import TranscriptCoverage
from ..polarity_score import polarity_score

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="coverage_features", description="Calculate coverage profile features")
    argparser.add_argument('coverage', metavar='coverage.bedgraph', help='Coverage data')
    argparser.add_argument('--prefix', default='', help='Prefix of property columns (to distinguish control and experiment)')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    coverages = TranscriptCoverage.each_in_file(args.coverage, header=False, dtype=int)
    with open_for_write(args.output_file) as output_stream:
        property_names = ['mean_coverage', 'total_coverage', 'polarity']
        prefixed_property_names = [f'{args.prefix}{name}' for name in property_names]
        header = ['transcript_id', *prefixed_property_names]
        print('\t'.join(header), file=output_stream)
        for transcript_coverage in coverages:
            coverage = transcript_coverage.coverage

            transcript_id = transcript_coverage.transcript_id
            mean_coverage = np.mean(coverage)
            total_coverage = np.sum(coverage)
            polarity = polarity_score(coverage)

            info = [transcript_id, mean_coverage, total_coverage, polarity]
            print('\t'.join(map(str, info)), file=output_stream)



import argparse
from ..gzip_utils import open_for_write
from ..utils import flatten
from ..tsv_reader import stream_table_column_highlighted

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="join_tables", description = "Join the second table to the first one")
    argparser.add_argument('reference_filename', metavar='reference_table.tsv', help='Table to be expanded in a tab-separated format')
    argparser.add_argument('reference_column', help='Name or indices of column in reference table to join')
    argparser.add_argument('supplementary_filenames', metavar='supplementary_tables.tsv', nargs='*', help='Supplementary tables in a tab-separated format')
    argparser.add_argument('--columns', nargs='*', dest='supplementary_columns', help='Names or indices of columns in supplementary tables to join')

    has_header_group = argparser.add_mutually_exclusive_group(required=True)
    has_header_group.add_argument('--header', action='store_true', dest='has_header', help="Table has header")
    has_header_group.add_argument('--no-header', action='store_false', dest='has_header', help="Tables doesn't have header")
    
    only_matching_group = argparser.add_mutually_exclusive_group(required=True)
    only_matching_group.add_argument('--only-matching', action='store_false', dest='allow_non_matching', help="Don't take lines which are present not in all files")
    only_matching_group.add_argument('--allow-non-matching', action='store_true', dest='allow_non_matching', help="Allow transcripts which are not present in CDS-annotation (they are not clipped)")

    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.supplementary_columns:
        supplementary_columns = args.supplementary_columns
    else:
        supplementary_columns = [args.reference_column] * len(args.supplementary_filenames)

    if len(args.supplementary_filenames) != len(supplementary_columns):
        raise ValueError(f'Number of columns ({args.supplementary_columns}) should match number of tables ({args.supplementary_filenames})')

    supplementary_tables = []
    for (table_filename, column) in zip(args.supplementary_filenames, supplementary_columns):
        table_stream = stream_table_column_highlighted(table_filename, [column], has_header=args.has_header, pop_column=True)
        if args.has_header:
            _, header = next(table_stream)
        mapping = dict(table_stream)
        supplementary_tables.append({'header': header, 'mapping': mapping, 'padding': [''] * len(header)})

    with open_for_write(args.output_file) as output_stream:
        table_stream = stream_table_column_highlighted(args.reference_filename, [args.reference_column], has_header=args.has_header, pop_column=False)
        if args.has_header:
            _, reference_header = next(table_stream)
            header = reference_header + flatten(tbl['header'] for tbl in supplementary_tables)
            print('\t'.join(header), file=output_stream)
        for (key, reference_row) in table_stream:
            if args.allow_non_matching:
                supplementary_rows = [tbl['mapping'].get(key, tbl['padding']) for tbl in supplementary_tables]
            else:
                supplementary_rows = [tbl['mapping'].get(key, None) for tbl in supplementary_tables]
                if not all(supplementary_rows):
                    continue
            row = reference_row + flatten(supplementary_rows)
            print('\t'.join(row), file=output_stream)



import argparse
import numpy as np
from ..gzip_utils import open_for_write
from ..tsv_reader import each_in_tsv

def sliding_row_with_window(arr, size):
    for idx in range(0, size // 2):
        yield (arr[idx], arr[0:size])
    for idx in range(size // 2, len(arr) - size // 2):
        yield (arr[idx], arr[(idx - size // 2):(idx - size // 2 + size)])
    for idx in range(len(arr) - size // 2, len(arr)):
        yield (arr[idx], arr[-size:])

# stddev --> 1
def standardize_stddev(val, val_mean, val_stddev):
    return val_mean + (val - val_mean) / val_stddev

# mean --> 0
def standardize_mean(val, val_mean, val_stddev):
    return val - val_mean

# mean --> 0
# stddev --> 1
# (aka z-score)
def standardize_zscore(val, val_mean, val_stddev):
    return (val - val_mean) / val_stddev

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="adjust", description = "Make length-dependend adjustment of features")
    argparser.add_argument('table', metavar='table.tsv', help='Table in tab-separated format')
    argparser.add_argument('--sort-field', dest='sorting_field', required=True, help='Field to sort a table')
    argparser.add_argument('--fields', nargs='*', dest='fields_to_correct', default=[], help='List of fields to correct')
    argparser.add_argument('--prefix', default='adjusted_', help='Prefix for corrected column name')
    argparser.add_argument('--window', metavar='SIZE', dest='window_size', type=int, default=100, help='Size of sliding window (default: %(default)s)')
    argparser.add_argument('--mode', choices=['zero_mean', 'unit_stddev', 'z-score'], default='z-score', help='How to standardize statistics (default: %(default)s)')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.mode == 'zero_mean':
        standardization = standardize_mean
    elif args.mode == 'unit_stddev':
        standardization = standardize_stddev
    elif args.mode == 'z-score':
        standardization = standardize_zscore
    else:
        raise ValueError(f'Unknown mode `{args.mode}`')

    dtype = float
    data = list(each_in_tsv(args.table))

    fields = list(data[0].keys())
    for field in [args.sorting_field, *args.fields_to_correct]:
        for row in data:
            row[field] = dtype(row[field])
    data.sort(key=lambda info: info[args.sorting_field])

    output_fields = fields[:]
    for field in args.fields_to_correct:
        output_fields.append(f'{args.prefix}{field}')

    with open_for_write(args.output_file) as output_stream:
        print('\t'.join(output_fields), file=output_stream)
        for row, window in sliding_row_with_window(data, args.window_size):
            modified_row = dict(row)
            for field in args.fields_to_correct:
                field_vals = [window_row[field] for window_row in window]
                modified_row[f'{args.prefix}{field}'] = standardization(row[field], val_mean=np.mean(field_vals), val_stddev=np.std(field_vals))
            print('\t'.join([str(modified_row[field]) for field in output_fields]), file=output_stream)



import dataclasses
from .dataclass_tsv_serializable import DataclassTsvSerializable
from ..profile_comparison import profile_difference_by_segment_counts, slope_by_segment_counts, segmentwise_sums

@dataclasses.dataclass(frozen=True)
class CoverageComparisonStats(DataclassTsvSerializable):
    transcript_id: str
    slope: float
    weighted_slope: float
    multipoint_slope: float
    logslope: float
    weighted_logslope: float
    multipoint_logslope: float
    profile_difference: float

    @classmethod
    def make_from_profiles(cls, transcript_id, control_profile, experiment_profile, segmentation):
        assert len(control_profile) == len(experiment_profile) == segmentation.segmentation_length
        control_sums = segmentwise_sums(segmentation, control_profile)
        experiment_sums = segmentwise_sums(segmentation, experiment_profile)

        slopes = slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=False)
        log_slopes = slope_by_segment_counts(control_sums, experiment_sums, segmentation, log_mode=True)
        info = {
            'transcript_id': transcript_id,
            'slope': slopes['center'],
            'weighted_slope': slopes['weighted_center'],
            'multipoint_slope': slopes['every_point'],
            'logslope': log_slopes['center'],
            'weighted_logslope': log_slopes['weighted_center'],
            'multipoint_logslope': log_slopes['every_point'],
            'profile_difference': profile_difference_by_segment_counts(control_sums, experiment_sums, segmentation),
        }
        return cls(**info)



import dataclasses
from .dataclass_tsv_serializable import DataclassTsvSerializable

# cds_start and cds_stop are 0-based related to transcript
@dataclasses.dataclass(frozen=True)
class CodingTranscriptInfo(DataclassTsvSerializable):
    gene_id: str
    transcript_id: str
    transcript_length: int
    cds_start: int = dataclasses.field(metadata={'converter': lambda x: int(x) if x != '' else None})
    cds_stop: int = dataclasses.field(metadata={'converter': lambda x: int(x) if x != '' else None})

    computable_properties = [('cds_length', 'cds_length'), ]

    @classmethod
    def load_transcript_cds_info(cls, cds_annotation_filename):
        transcript_infos = cls.each_in_file(cds_annotation_filename)
        return {tr_info.transcript_id: tr_info  for tr_info in transcript_infos}

    @property
    def cds_length(self):
        if self.is_coding:
            return self.cds_stop - self.cds_start
        else:
            return None
    
    def cds_profile(self, profile):
        return profile[self.cds_start : self.cds_stop]

    @property
    def is_coding(self):
        return self.cds_start and self.cds_stop



import dataclasses
from typing import Union
from decimal import Decimal
from .dataclass_tsv_serializable import DataclassTsvSerializable

@dataclasses.dataclass
class CoverageInterval(DataclassTsvSerializable):
    chrom: str
    start:int
    stop: int
    coverage: Union[float, int] = dataclasses.field(metadata={'skip_conversion': True})
    dtype: dataclasses.InitVar[type] = float
    def __post_init__(self, dtype):
        try:
            self.coverage = dtype(self.coverage)
        except:
            # if dtype is integer, string "1.2345e6" can't be directly converted to int
            self.coverage = dtype(Decimal(self.coverage))



import argparse
import numpy as np
from ..utils import align_iterators, pool_profiles, get_constant_intervals
from ..dto.coverage_interval import CoverageInterval
from ..dto.transcript_coverage import TranscriptCoverage
from ..gzip_utils import open_for_write

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="pool_coverage", description="Pool coverage profiles")
    argparser.add_argument('coverages', nargs='*', help='Alignment coverages in bedgraph format')
    argparser.add_argument('--only-matching', action='store_true', help="Don't pool coverages of transcripts which are present not in all files")
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    argparser.add_argument('--output-mode', choices=['sum', 'mean'], default='sum', help="What to report")
    argparser.add_argument('--dtype', choices=['int', 'float'], default='int', help="Make int or float-valued coverage (default: int)")
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if args.dtype == 'int':
        dtype = int
    elif args.dtype == 'float':
        dtype = float
    else:
        raise ValueError('dtype should be either int or float')

    coverage_streams = [TranscriptCoverage.each_in_file(coverage_fn, header=False, dtype=dtype) for coverage_fn in args.coverages]

    with open_for_write(args.output_file) as output_stream:
        aligned_transcripts = align_iterators(coverage_streams, key=lambda transcript_coverage: transcript_coverage.transcript_id, check_sorted=True)
        for (transcript_id, coverages) in aligned_transcripts:
            if args.only_matching and not all(coverages):
                continue
            coverages = filter(lambda x: x, coverages)
            coverages = [transcript_coverage.coverage for transcript_coverage in coverages]
            if args.output_mode == 'sum':
                pooled_coverage_profile = np.sum(coverages, axis=0)
            elif args.output_mode == 'mean':
                pooled_coverage_profile = np.mean(coverages, axis=0)
            else:
                raise ValueError(f'Unknown output_mode `{args.output_mode}`')
            pooled_coverage_intervals = get_constant_intervals(pooled_coverage_profile)
            pooled_bedgraph = (CoverageInterval(transcript_id, *interval, dtype=dtype) for interval in pooled_coverage_intervals)
            CoverageInterval.print_tsv(pooled_bedgraph, header=False, file=output_stream)



import argparse
import seaborn as sns
import matplotlib.pyplot as plt
from ..tsv_reader import each_in_tsv

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="plot_distribution", description="Plot distributions of a property or several properties")
    argparser.add_argument('table', metavar='table.tsv', help='Table in tab-separated format')
    argparser.add_argument('--fields', nargs='*', required=True, help='Fields to plot on the same figure')
    argparser.add_argument('--labels', nargs='*', help='Set legend labels (by default field names are used)')
    argparser.add_argument('--title', help="Add plot title")
    argparser.add_argument('--xlim', nargs=2, type=float, help="Add limits for X-axis (in form `--xlim min max`)")
    argparser.add_argument('--ylim', nargs=2, type=float, help="Add limits for Y-axis (in form `--ylim min max`)")
    argparser.add_argument('--zero-line', help="Add vertical line at zero of specified color")

    has_legend_group = argparser.add_mutually_exclusive_group(required=True)
    has_legend_group.add_argument('--legend', action='store_true', dest='has_legend', help="Table has legend")
    has_legend_group.add_argument('--no-legend', action='store_false', dest='has_legend', help="Tables doesn't have legend")
    
    argparser.add_argument('--display', action='store_true', help="Display resulting figure")
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    dtype = float
    data = list(each_in_tsv(args.table))

    fields = args.fields

    if len(fields) == 0:
        raise ValueError(f'Specify at least one field to plot')

    if args.labels:
        if len(fields) != len(args.labels):
            raise ValueError(f'Number of labels ({len(args.labels)}) should be equal to number of fields ({len(fields)})')
        labels = args.labels
    else:
        labels = fields

    for field in fields:
        for row in data:
            row[field] = dtype(row[field])

    plt.figure()

    if args.title:
        plt.title(args.title)
    
    for (field, label) in zip(fields, labels):
        values = [row[field] for row in data]
        sns.kdeplot(values, label = field)

    if args.xlim:
        plt.xlim(*args.xlim)

    if args.ylim:
        plt.ylim(*args.ylim)

    if args.zero_line:
        plt.axvline(x=0, color=args.zero_line)

    if args.output_file:
        plt.savefig(args.output_file)

    if args.display:
        plt.show()


import sys
import argparse
from ..gzip_utils import open_for_write
from ..dto.interval import Interval
from ..dto.coding_transcript_info import CodingTranscriptInfo
from ..clipping import Clipper

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(
            prog = "clip_cds",
            description = "Clip any bed file in transcriptomic coordinates to CDS-region",
        )
    argparser.add_argument('cds_annotation', metavar='cds_annotation.tsv', help='CDS annotation') # 'gencode.vM22.cds_features.tsv'
    argparser.add_argument('bedfile', metavar='bedfile.bed', help = 'Coverage or segmentation file in bed format (3 standard columns + any number of non-standard)')

    # start and stop codon can have piles of reads, so we usually want to drop them
    # flank lengths to drop are of 15nt ~= half-ribosome (~half of riboseq footprint length)
    argparser.add_argument('--drop-5-flank', metavar='N', type=int, default=0, help="Clip N additional nucleotides from transcript start (5'-end)")
    argparser.add_argument('--drop-3-flank', metavar='N', type=int, default=0, help="Clip N additional nucleotides from transcript end (3'-end)")

    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    argparser.add_argument('--allow-non-matching', action='store_true', help="Allow transcripts which are not present in CDS-annotation (they are not clipped)")
    argparser.add_argument('--contig-naming', dest='contig_naming_mode', choices=['original', 'window'], default='window', help="Use original (chr1) or modified (chr1:23-45) contig name for resulting intervals")
    return argparser

def main():
    argparser = configure_argparser()
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    if (args.allow_non_matching) and (args.contig_naming_mode != 'original'):
        print('Attention! When `--allow-non-matching` is set, only `--contig-naming original` will give consistent contig names', file=sys.stderr)

    cds_info_by_transcript = CodingTranscriptInfo.load_transcript_cds_info(args.cds_annotation)
    bed_stream = Interval.each_in_file(args.bedfile)
    clipper = Clipper(contig_naming_mode=args.contig_naming_mode,
                      drop_5_flank=args.drop_5_flank,
                      drop_3_flank=args.drop_3_flank)
    with open_for_write(args.output_file) as output_stream:
        for interval in clipper.bedfile_clipped_to_cds(bed_stream, cds_info_by_transcript, allow_non_matching=args.allow_non_matching):
            print(interval, file=output_stream)



def polarity_score(coverage):
    total_coverage = 0
    numerator = 0
    for i in range(len(coverage)):
        total_coverage = total_coverage + coverage[i]
        numerator += coverage[i] * i
    if total_coverage == 0:
        return 0
    polarity_score = numerator / total_coverage
    normalized_score = (polarity_score * 2 / (len(coverage) - 1)) - 1
    return normalized_score



import pprint
import six
import subprocess
import pybedtools
from pybedtools.logger import logger
from pybedtools.helpers import BUFSIZE, BEDToolsError
from pybedtools import BedTool

def coreutils_sort(self, **kwargs):
    if 'instream' not in kwargs:
        kwargs['instream'] = self.fn

    cmds, tmp, stdin = self.handle_coreutils_sort_kwargs(prog='sort', **kwargs)
    stream = call_coreutils_sort(cmds, tmp, stdin=stdin)
    return BedTool(stream)

def handle_coreutils_sort_kwargs(self, prog='sort', instream=None, **kwargs):
    """
    Handle coreutils sort program calls.

    *kwargs* are passed directly from the calling method (self.coreutils_sort).

    This method figures out, given how this BedTool was constructed, what
    to send to BEDTools programs -- for example, an open file to stdin with
    the `-` argument, or a filename with the `-a` argument.

    *instream* can be e.g., self.fn or 'a.bed' or an iterator.
    """
    pybedtools.logger.debug(
        'BedTool.handle_coreutils_sort_kwargs() got these kwargs:\n%s',
        pprint.pformat(kwargs))

    stdin = None

    # Decide how to send instream to sort.
    # If it's a BedTool, then get underlying stream
    if isinstance(instream, BedTool):
        instream = instream.fn

    # Filename? No pipe, just provide the file
    if isinstance(instream, six.string_types):
        stdin = None
        input_fn = instream
    # A generator or iterator: pipe it as a generator of lines
    else:
        stdin = (str(i) for i in instream)
        input_fn = '-'

    # If stream not specified, then a tempfile will be created
    if kwargs.pop('stream', None):
        tmp = None
    else:
        output = kwargs.pop('output', None)
        if output:
            tmp = output
        else:
            tmp = BedTool._tmp()

    additional_args = kwargs.pop('additional_args', None)

    # Parse the kwargs into BEDTools-ready args
    cmds = [prog]

    for key, value in sorted(list(kwargs.items()), reverse=True):
        if isinstance(value, bool):
            if value:
                cmds.append('--' + key)
            else:
                continue
        elif isinstance(value, list) or isinstance(value, tuple):
            value = list(map(str, value))

            # sort --key 1,1 --key 2,2r -k 5,5
            for val in value:
                if len(key) == 1:
                    cmds.append('-' + key)
                else:
                    cmds.append('--' + key)
                cmds.append(str(val))
        else:
            cmds.append('--' + key)
            cmds.append(str(value))

    if additional_args:
        cmds.append(additional_args)

    cmds.append(input_fn)
    return cmds, tmp, stdin



def call_coreutils_sort(cmds, tmpfn=None, stdin=None):
    """
    Use subprocess.Popen to call BEDTools and catch any errors.

    Output goes to *tmpfn*, or, if None, output stays in subprocess.PIPE and
    can be iterated over.

    *stdin* is an optional file-like object that will be sent to
    subprocess.Popen.

    Prints some useful help upon getting common errors.

    """
    input_is_stream = stdin is not None
    output_is_stream = tmpfn is None

    try:
        # coming from an iterator, sending as iterator
        if input_is_stream and output_is_stream:
            logger.debug('helpers.call_coreutils_sort(): input is stream, output is stream')
            logger.debug('helpers.call_coreutils_sort(): cmds=%s', ' '.join(cmds))
            p = subprocess.Popen(cmds,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 stdin=subprocess.PIPE,
                                 bufsize=BUFSIZE)
            for line in stdin:
                p.stdin.write(line.encode())
            p.stdin.close()  # This is important to prevent deadlocks

            output = (i.decode('UTF-8') for i in p.stdout)
            stderr = None

        # coming from an iterator, writing to file
        if input_is_stream and not output_is_stream:
            logger.debug('helpers.call_coreutils_sort(): input is stream, output is file')
            logger.debug('helpers.call_coreutils_sort(): cmds=%s', ' '.join(cmds))
            outfile = open(tmpfn, 'wb')
            p = subprocess.Popen(cmds,
                                 stdout=outfile,
                                 stderr=subprocess.PIPE,
                                 stdin=subprocess.PIPE,
                                 bufsize=BUFSIZE)
            if hasattr(stdin, 'read'):
                stdout, stderr = p.communicate(stdin.read())
                p.stdin.close()
            else:
                for item in stdin:
                    p.stdin.write(item.encode())
                stdout, stderr = p.communicate()
                p.stdin.close()
            output = tmpfn
            outfile.close()

        # coming from a file, sending as iterator
        if not input_is_stream and output_is_stream:
            logger.debug('helpers.call_coreutils_sort(): input is filename, output is stream')
            logger.debug('helpers.call_coreutils_sort(): cmds=%s', ' '.join(cmds))
            p = subprocess.Popen(cmds,
                                 stdout=subprocess.PIPE,
                                 stderr=subprocess.PIPE,
                                 bufsize=BUFSIZE)
            output = (i.decode('UTF-8') for i in p.stdout)
            stderr = None

        # file-to-file
        if not input_is_stream and not output_is_stream:
            logger.debug('helpers.call_coreutils_sort(): input is filename, output is filename (%s)', tmpfn)
            logger.debug('helpers.call_coreutils_sort(): cmds=%s', ' '.join(cmds))
            outfile = open(tmpfn, 'wb')
            p = subprocess.Popen(cmds,
                                 stdout=outfile,
                                 stderr=subprocess.PIPE,
                                 bufsize=BUFSIZE)
            stdout, stderr = p.communicate()
            output = tmpfn
            outfile.close()

        if stderr:
            if isinstance(stderr, bytes):
                stderr = stderr.decode('UTF_8')
            raise BEDToolsError(subprocess.list2cmdline(cmds), stderr)

    except (OSError, IOError) as err:
        print('%s: %s' % (type(err), os.strerror(err.errno)))
        print('The command was:\n\n\t%s\n' % subprocess.list2cmdline(cmds))
        raise err

    return output

pybedtools.BedTool.coreutils_sort = coreutils_sort
pybedtools.BedTool.handle_coreutils_sort_kwargs = handle_coreutils_sort_kwargs



import argparse
from ..gzip_utils import open_for_write
from ..annotation import Annotation
from ..dto.coding_transcript_info import CodingTranscriptInfo

def parse_condition(condition_str):
    k,vs = condition_str.split('=', maxsplit=1)
    return (k, vs.split(','))

# Known issue: filters now treat all attribute values as strings
def create_record_filter(condition_config):
    k,vs = condition_config
    return lambda rec: str(rec.attributes.get(k, '')) in vs

def configure_argparser(argparser=None):
    if not argparser:
        argparser = argparse.ArgumentParser(prog="extract_cds_annotation", description = "Extract CDS annotation in transcriptomic coordinates from genomic annotation")
    argparser.add_argument('gtf_annotation', metavar='annotation.gtf', help='Genomic annotation in GTF-format')
    argparser.add_argument('--output-file', '-o', dest='output_file', help="Store results at this path")
    argparser.add_argument('--attr-filter', action='append', dest='filters', default=[], 
                                            help="Filter records so that attributes has one of specified values.\n"
                                                 "Format: `attribute=value_1,value_2,...`")
    # Known issue: Right now all multivalue keys are ignored, so filters won't work with them
    # ToDo: add ability to specify multivalue keys and add filters like `--filter-has tag value`
    return argparser

def main():
    argparser = configure_argparser(argparser)
    args = argparser.parse_args()
    invoke(args)

def invoke(args):
    condition_configs = [parse_condition(condition_str) for condition_str in args.filters]
    filters = [create_record_filter(condition_config) for condition_config in condition_configs]
    filter_conjuction = lambda rec: all(f(rec) for f in filters)
    relevant_attributes = {k for (k,v) in condition_configs}

    annotation = Annotation.load(
        args.gtf_annotation,
        relevant_attributes=relevant_attributes,
        multivalue_keys=set(),
        ignore_unknown_multivalues=True,
        condition=filter_conjuction,
    )
    with open_for_write(args.output_file) as output_stream:
        print(CodingTranscriptInfo.header(), file=output_stream)
        for transcript_id in annotation.transcript_by_id:
            cds_info = annotation.coding_transcript_info(transcript_id)
            print(cds_info, file=output_stream)



import dataclasses
from itertools import groupby
from .dto.interval import Interval

def segments_clipped_to_window(segments, window_start, window_stop, contig_name):
    for segment in segments:
        segment_start = max(0, segment.start - window_start)
        segment_stop = min(window_stop, segment.stop) - window_start
        if segment_stop - segment_start > 0:
            yield Interval(contig_name, segment_start, segment_stop, segment.rest)

@dataclasses.dataclass
class Clipper:
    contig_naming_mode: str = 'window'
    drop_5_flank: int = 0
    drop_3_flank: int = 0

    def __post_init__(self):
        if self.contig_naming_mode not in ['original', 'window']:
            raise ValueError(f'Unknown contig naming mode `{self.contig_naming_mode}`')

    def customize_contig_name(self, contig_name, window_start, window_stop):
        if self.contig_naming_mode == 'original':
            return contig_name
        elif self.contig_naming_mode == 'window':
            return f'{contig_name}:{window_start}-{window_stop}'
        else:
            raise ValueError(f'Unknown contig naming mode `{self.contig_naming_mode}`')

    def segments_clipped_to_cds(self, segments, cds_info, contig_name):
        window_start = cds_info.cds_start + self.drop_5_flank
        window_stop = cds_info.cds_stop - self.drop_3_flank
        customized_contig_name = self.customize_contig_name(contig_name, window_start, window_stop)
        yield from segments_clipped_to_window(segments, window_start, window_stop, customized_contig_name)

    def bedfile_clipped_to_cds(self, bed_stream, cds_info_by_transcript, allow_non_matching=False):
        '''
        Attention! if `contig_naming_mode` is not "original" and `allow_non_matching` is True,
        then contig names will be in inconsistent formats:
        transcripts with have matching cds_info, will have customized contig names,
        while transcripts without matching cds_info will have original contig names.
        '''
        for (contig_name, segments) in groupby(bed_stream, lambda segment: segment.chrom):
            if contig_name in cds_info_by_transcript:
                cds_info = cds_info_by_transcript[contig_name]
                if cds_info.is_coding:
                    yield from self.segments_clipped_to_cds(segments, cds_info, contig_name)
                elif allow_non_matching:
                    yield from segments
            else:
                if allow_non_matching:
                    yield from segments



import setuptools
from io import open
import os

version_module = {}
dirname = os.path.dirname(__file__)
with open(os.path.join(dirname, "src/papolarity/version.py")) as fp:
    exec(fp.read(), version_module)
    __version__ = version_module['__version__']

with open(os.path.join(dirname, "README.md"), encoding='utf8') as fh:
    long_description = fh.read()

setuptools.setup(
    name="papolarity",
    version=__version__,
    author="Ilya Vorontsov",
    author_email="vorontsov.i.e@gmail.com",
    description="Papolarity is a tool to analyze polarity of transcriptomic alignments such as Ribo-seq and RNA-seq.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/autosome-ru/papolarity",
    license="Papolarity is licensed under WTFPL, but if you prefer more standard licenses, feel free to treat it as MIT license",
    packages=setuptools.find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    keywords='bioinformatics NGS coverage alignment polarity Ribo-seq RNA-seq',
    classifiers=[
        "Topic :: Scientific/Engineering :: Bio-Informatics",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 5 - Production/Stable",
        "Operating System :: OS Independent",
        'Programming Language :: Python :: 3.7',
    ],
    python_requires='>=3.7, <3.8',
    install_requires=['pybedtools', 'numpy >= 1.8.0', 'pasio >= 1.1.2', 'sklearn', 'six', 'matplotlib', 'seaborn'],
    extras_require={
        'dev': ['pytest', 'pytest-benchmark', 'flake8', 'tox', 'wheel', 'twine', 'setuptools_scm'],
    },
    entry_points={
        'console_scripts': [
            'papolarity=papolarity.cli:main',
        ],
    },
    use_scm_version=False,
    setup_requires=['setuptools_scm'],
)



import itertools
import dataclasses
from typing import List
import numpy as np
import pasio
from .utils import pool_profiles
from .dto.interval import Interval

def make_joint_segmentation(coverages, splitter):
    pooled_coverage = pool_profiles(coverages)
    if len(pooled_coverage) == 0:
        return None
    return list(pasio.segments_with_scores(pooled_coverage, splitter))

def stabilize_profile(profile, segments):
    stable_profile = np.zeros_like(profile)
    for segment in segments:
        start = segment.start
        stop = segment.stop
        stable_profile[start:stop] = np.mean(profile[start:stop])
    return stable_profile

@dataclasses.dataclass(frozen=True)
class Segmentation:
    chrom: str
    segments: List[Interval] # bed-coordinates [a, b)

    def __post_init__(self):
        object.__setattr__(self, 'segments', sorted(self.segments)) # assign frozen attribute.
        if any(self.chrom != segment.chrom for segment in self.segments):
            raise ValueError(f"Segmentation intervals all should be on specified chromosome/contig `{self.chrom}`")
        if len(self.segments) > 0 and self.segments[0].start != 0:
            raise ValueError(f"Segmentation should cover entire interval (but first segment {self.segments[0]} doesn't start from zero)")
        if any(self.segments[i].stop != self.segments[i + 1].start for i in range(len(self.segments) - 1)):
            raise ValueError(f"Segments should be contigious but weren't: {self.segments}")

    @property
    def segmentation_length(self):
        return self.segments[-1].stop

    @property
    def num_segments(self):
        return len(self.segments)

    def stabilize_profile(self, profile):
        stable_profile = np.zeros_like(profile)
        for segment in self.segments:
            start = segment.start
            stop = segment.stop
            stable_profile[start:stop] = np.mean(profile[start:stop])
        return stable_profile

    # clip flanks so that region is in specified window
    def clip_to_window(self, start, stop):
        return self.clip_flanks(start, self.segmentation_length - stop)

    # clip flanks of specified length
    def clip_flanks(self, flank_5, flank_3):
        clip_len = self.segmentation_length - flank_3 - flank_5
        clipped_segments = []
        for segment in self.segments:
            segment_start = max(0, segment.start - flank_5)
            segment_stop = min(clip_len, segment.stop - flank_5)
            if (segment_stop <= 0) or (clip_len <= segment_start):
                continue
            clipped_segment = Interval(self.chrom, segment_start, segment_stop)
            clipped_segments.append(clipped_segment)
        return Segmentation(self.chrom, clipped_segments)

    @classmethod
    def each_in_file(cls, filename, force_gzip=None, header=False):
        segment_stream = Interval.each_in_file(filename, header=header, force_gzip=force_gzip)
        for (chrom, segments_iter) in itertools.groupby(segment_stream, key=lambda segment: segment.chrom):
            segments = list(segments_iter)
            yield cls(chrom, segments)
